/*
 *
 * Modification History:
 *
 * March 30, 1993 by Robin Miller, MILLER W J & ASSOCIATES
 *	Fixed problem where the local tty speed was inadvertantly
 *	set to the modem speed, when not using stdin, since only
 *	one tty params structure was being used (my bug).  This
 *	caused garbage to appear on the local terminal during the
 *	file transfer, since the baud rate was wrong.
 *
 * January 11, 1993 by Robin Miller, MILLER W J & ASSOCIATES
 *	Ported 'qcp' to DEC OSF/1 Operating System.
 *	o  added functions to setup terminal attributes.
 *	o  added debugging option "-D debug-log".
 */
/*				11-Dec-89 10:27:59am										*/

/*--------------------------------------------------------------------------*/
/*  History: main.c, V0.0, ??-Nov-86 11:36:31am, Dan Dodge,	Baseline		*/
/*  				 V0.1, 13-Jan-87 11:36:31am, Dan Hildebrand				*/
/*----------------------------------------------------------------------------

     This file contains the command line parser and some of  the  commonly
     needed  routines  for  the various protocols that make up the Quantum
     Communications Protocol family. User written protocols should link in
     this module and include "main.h"  to  interface  cleanly  to  Quantum
     communications products.    Use  of  this command line interface also
     makes user written protocols  installable  into  the  /qterm/protocol
     file, thus  making  the  protocol  usable  from  within  Qterm. Other
     application programs ( such as bulletin board programs ) can refer to
     the /qterm/protocol file when a choice of protocol is needed.

     The  following  routines  must be defined and linked with this source
     module to produce a working protocol.    Please  refer  to  the  file
     "protocol.doc" for details.

     void help(void);				- Display help message
     int extra_option(char *);		- Process extra command line options
     void init_mdm(void);			- Initialize protocol variables
     void multi_send_check(void);	- Check if batch mode supported
     void done(int );				- Exit with status, shutdown protocol
     void final_init(void);			- Last init called before snd/rcv
     int snd(char *);				- Send a file
     receive_files( int argc, char *argv[] ) - Receive one or more files

----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>
#if defined(_QNX_SOURCE)
#  include <sys/fsys.h>
#  include <sys/dev.h>
#else /* !defined(_QNX_SOURCE) */
#  include <setjmp.h>
#  include <signal.h>
#  include <sys/ioctl.h>
#  include <sys/time.h>
#endif /* defined(_QNX_SOURCE) */
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <string.h>
#if defined(_QNX_SOURCE)
#  include <i86.h>
#endif /* defined(_QNX_SOURCE) */
/* Give EXT a null definition so that externals become local to this module	*/
#define	EXT
#include "main.h"

unsigned ungot_char = 0xffff;

#if defined(_QNX_SOURCE)

unsigned old_mode;

char device[ MAX_TTY_NAME+1 ];

#else /* !defined(_QNX_SOURCE) */

/*
 * Generic return codes for library functions/system calls.
 */
#define SUCCESS				0
#define FAILURE				-1
#define WARNING				1

extern int	getopt(int, char *[],char *);
extern char	*optarg;
extern int	optind, opterr, optopt;

char *device;

struct tty_params {
	int	tty_saved;		/* tty characteristics saved.	*/
	int	tty_ldisc;		/* The saved line discipline.	*/
	struct termios tty_smodes;	/* The saved terminal modes.	*/
	struct termios tty_cmodes;	/* The current terminal modes.	*/
} local_tty, modem_tty;

/*
 * Non-QNX O/S Forward References.
 */
void	a_abrt(int err, char *fn, int fd);
ssize_t	dev_read (int fd, void *buf, ssize_t n, unsigned minc, unsigned time,
				unsigned timeout, pid_t proxy, int *armed);
ssize_t	a_read(int fildes, void *buf, size_t nbyte);
void	timerALRM(int);
int	delay(int usecdelay);
int	dev_ischars(int fd);
int	save_tty(int fd, struct tty_params *tp);
int	restore_tty(int fd, struct tty_params *tp);
int	setup_tty(int fd, struct tty_params *tp);
void	print_usage( char *argv[] );

static	jmp_buf timeoutbuf, alrmbuf;

#endif /* defined(_QNX_SOURCE) */

void
main( int argc, char *argv[] ) {
	int file_count = 0;
#if defined(_QNX_SOURCE)
	struct _dev_info_entry dinfo;
#endif /* defined(_QNX_SOURCE) */
	int opt;

#if defined(_QNX_SOURCE)
	if ( need_usage( argv ) || argc < 2 )
		print_usage( argv );
#else /* defined(_QNX_SOURCE) */
	optind = 0;
	if ( argc < 2 )
		print_usage( argv );
#endif /* defined(_QNX_SOURCE) */

	prefix[0] = '\0';	/* No prefix and forced file name off by default	*/
	logfile[0] = '\0';	/* No activity log file by default */
	force = newer = relaxed = today = verbose = quiet = qterm = FALSE;
	make_dir = TRUE;
	onesec = ONESEC;
	retry_limit = RETRY_MAX;

	++optind;		/* Skip over command name	*/
	/*-- Check for a device over-ride ----*/
	if ( *argv[ optind ] == '/' ) {
		if ( (mdm = open( argv[1], O_RDWR ) ) == -1) {
			fprintf( stderr, "qcp: unable to open device '%s'.\r\012", argv[1] );
			exit( -1 );
			}
		++optind;		/* Skip over specified device name	*/
		}

	/*-- No device override so use stdin/stdout to communicate ----*/
	else {
#if defined(_QNX_SOURCE)
		if ( dev_info( 0, &dinfo ) == -1 ) {
			fprintf( stderr, "qcp: Stdin not a device.\r\012" );
			exit( -1 );
			}
		strncpy( device, dinfo.tty_name, MAX_TTY_NAME );
		if ( (mdm = open( device, O_RDWR ) ) == -1) {
			fprintf( stderr, "qcp: unable to open stdin as r/w.\r\012" );
			exit( -1 );
			}
		use_stdin = TRUE;
#else /* !defined(_QNX_SOURCE) */
		if ( (device = ttyname(fileno(stdin))) == NULL ) {
			fprintf( stderr, "qcp: Stdin not a device.\r\012" );
			exit( -1 );
		}
		if ( (mdm = open( device, O_RDWR ) ) == -1) {
			fprintf( stderr, "qcp: unable to open stdin as r/w.\r\012" );
			exit( -1 );
			}
		use_stdin = TRUE;
#endif /* defined(_QNX_SOURCE) */
	}

	init_mdm();			/* Setup the selected modem port and initialize		*/
						/*	any variables associated with the protocol.		*/

	if		( ( *argv[optind] | ' ' ) == 'r')					sending = FALSE;
	else if ( ( *argv[optind] | ' ' ) == 's'  &&  argc > 2 )	sending = TRUE;
	else {
#if defined(ultrix)
		fprintf( stderr, "Invalid argument '%s', must be 'se' or 're'.\r\012",
#else /* !defined(ultrix) */
		fprintf( stderr, "Invalid argument '%s', must be 'se' or 're'.\r\l",
#endif /* defined(ultrix) */
						argv[optind] );
		restore_opt( mdm );
		exit( -1 );
		}
	++optind;

	/*---- Ready for command line options ----*/
	while( ( opt = getopt( argc, argv, "mnrtuvVQf:l:p:s:D:" ) ) != -1 ) {
		switch( opt ) {
			case 'f':		strcpy( prefix, optarg );
							force = TRUE;				break;
			case 'l':		strcpy( logfile, optarg );	break;
			case 'm':		make_dir = FALSE;			break;
			case 'n':		newer = TRUE;				break;
			case 'r':		relaxed = TRUE;
							onesec <<= 2;		/* Timeouts 4 times larger	*/
							retry_limit <<= 1;	/* Twice as many retries	*/
							break;
			case 's':		sub_packet_size = atoi( optarg );
							sub_packet_size >>= 3;	/* Divide by 8 for sub_packet	*/
							if ( !sub_packet_size ) {
								fprintf( stderr, "qcp: Packet size too small.\r\012" );
								exit( -1 );
								}
							if ( sub_packet_size > MAXSIZE ) {
								fprintf( stderr, "qcp: Packet size too large. Limit is %d.\r\012",
											MAXSIZE*8);
								exit( -1 );
								}
							break;
			case 't':		today = TRUE;					break;
			case 'u':		unlink_file = TRUE;				break;
			case 'V':
			case 'v':		verbose = TRUE;					break;
			case 'q':		quiet = TRUE;					break;
			case 'Q':		qterm = TRUE;					break;
			case 'p':		strcpy( prefix, optarg );		break;
			case '?': 		if ( !extra_option( (char *)&optopt ) ) {
								fprintf( stderr, "%s : Unknown option\r\012", optopt );
								exit( -1 );
								}
							break;
			case 'D':	if (freopen (optarg, "a", stderr) == NULL) {
					    perror("freopen");
					    restore_opt( mdm );
					    exit ( -1 );
					}
					debug_flag = TRUE;
					break;
			default: 		++file_count;		/* Count filenames		*/
							continue;			/*	but don't null them	*/
			}
		}

	/*
	 * Make stderr buffered, so timing is not affected by output.
	 */
	if (debug_flag) {
	    if ((debug_buffer = (char *) malloc (BUFSIZ)) == NULL) {
		perror("malloc");
		restore_opt( mdm );
		exit ( -1 );
	    } else {
		setbuf (stderr, debug_buffer);
	    }
	}

	if ( file_count > 1 ) multi_send_check();	/* Check if multi-send OK	*/

	final_init();						/* Last changes before starting		*/

	if ( !use_stdin ) raw( 0 );		/* Raw console input for abort requests	*/
	if ( sending )	send_files(		argc - optind, argv + optind );
	else	 		receive_files(	argc - optind, argv + optind );

	done( 0 );				/* Protocol dependant termination routine		*/
	}

void
send_files( int argc, char *argv[] ) {
	FILE *fp;
	int i, rc;
	char	xfile_buf[110];	/* Place to read file name entries from xfile	*/

	for( i = 0; i < argc; ++i ) {
		if ( argv[i][0] == 'x'  &&  argv[i][1] == '=' ) {
			multi_send_check();		/* Does the protocol do multi-send ?	*/
			if ( !(fp = fopen( &argv[i][2], "r" ) ) ) {
				fprintf( stderr, "Unable to open index '%s'.\r\012", &argv[i][2] );
				exit( -1 );
				}
			while( fgetln( xfile_buf, 100, fp ) ) {
				rc = snd( xfile_buf );
				writelog( "SND", rc, xfile_buf );
				if( !rc )	done( -1 );
				}
			fclose( fp );
			}
		else {
			rc = snd( argv[i] );
			writelog( "SND", rc, argv[i] );
			if( !rc )	done( -1 );
			}
		}
	done( TRUE );
	}

read_mdm( int nchars, int timeout, uchar *buf ) {
	ssize_t n, i;
	int ungot = 0;

	if (debug_flag) {
#if defined(_QNX_SOURCE)
	    putdbg("read_mdm( int nchars = %d, int timeout = %d, uchar *buf = %#x )\n",
#else /* !defined(_QNX_SOURCE) */
	    putdbg("read_mdm( int nchars = %d, int timeout = %d, uchar *buf = %#lx )\n",
#endif /* defined(_QNX_SOURCE) */
							nchars, timeout, buf);
	}
	if ( ungot_char != 0xffff ) {		/* Process unget buffer first	*/
		--nchars;
		*buf++ = (uchar) ungot_char;
		ungot_char = 0xffff;
		ungot = 1;
		}
	if ( timeout == -1 ) timeout = 0;
/*	n = dev_read( mdm, buf, nchars, nchars, 0, timeout, 0, NULL );	*/
	n = 0;
	while ( nchars  &&
			(i = dev_read( mdm, buf, nchars, nchars, 0, timeout, 0, 0 ) ) ) {
		if ( i == -1 ) perror( "read_mdm()" );
		buf += i;
		nchars -= i;
		n += i;
		}
	if (debug_flag) {
	    dump( "data returned", (n + ungot), (buf - (n + ungot) ));
	}
	return( n + ungot );
	}

void
writelog( char *str, int rc, char *msg ) {
	int i = 0;
	time_t tt;
	FILE *fp;

	if ( logfile[0] && *msg ) {
		time( &tt );
		while( i++ < 5 ) {	/* Try 5 times to open the potentially busy file*/
			if ( !( fp = fopen( logfile, "a" ) ) ) {
				wait_ticks( 10 );			/* 0.5 seconds between attempts	*/
				continue;
				}
			fprintf( fp, "%s  %s %d  %s\r\012", ctime( &tt ), str, rc, msg );
			fclose( fp );
			break;
			}
		if ( i == 5 )
			fprintf( stderr, "Unable to append to Logfile '%s'.\r\012", logfile );
		}
	}

void
write_mdm( unsigned nchars, void *buf ) {
	if (debug_flag) {
#if defined(_QNX_SOURCE)
	    putdbg("write_mdm( unsigned nchars = %d, void *buf = %#x )\n", nchars, buf);
#else /* !defined(_QNX_SOURCE) */
	    putdbg("write_mdm( unsigned nchars = %d, void *buf = %#lx )\n", nchars, buf);
#endif /* defined(_QNX_SOURCE) */
	}
	write( mdm, buf, nchars );
	if (debug_flag) {
	    dump("data written", nchars, buf);
	}
}

void
putdbg( char *fmt, ... )
{
	va_list	arglist;

	if ( !debug_flag ) return;
	va_start( arglist, fmt );
	vfprintf( stderr, fmt, arglist );
	fflush( stderr );
}

void
putmsg( char *fmt, ... ) {
	va_list	arglist;

	if ( quiet  ||  use_stdin ) return;
	va_start( arglist, fmt );
	vprintf( fmt, arglist );
	fflush( stdout );
	}

void
puterr( char *fmt, ... ) {
	va_list arglist;

	if ( debug_flag ) {
	    va_start( arglist, fmt );
	    vfprintf( stderr, fmt, arglist );
	    fputs("\n", stderr);
	    fflush( stderr );
	} else {
	    if ( quiet  ||  use_stdin  ||  !verbose ) return;
	    va_start( arglist, fmt );
	    vprintf( fmt, arglist );
	    fflush( stdout );
	}
}

void
wait_ticks( int time ) {
	if ( time ) delay( time * 50 );		/* Delay units of 50 milliseconds 	*/
	}

abort_check() {
	char c;

	if ( use_stdin  ||  !dev_ischars( 0 ) ) return( FALSE );
	read( 0, &c, 1 );
	if ( c == ' ' || c == 27 || c == '\r' ) {
		printf( "\r\012Stop %s (y/n) ? ", sending ? "sending" : "receiving" );
		fflush( stdout );
		read( 0, &c, 1 );
		printf( "\r\012" );
		fflush( stdout );
		return( (c == 'y')  ||  (c == 'Y') );
		}
	return( FALSE );
	}

/*----------------------------------------------------------------------------
     This  routine  translates a received filename into a filename allowed
     by the current command line options. If the "-f forced_filename"  option
     is set,  the  specified  filename  will  be  placed into "new".  If a
     filename prefix was specified with the -p option, any node and  drive
     information  on  the  received  filename  will  be  removed  and  the
     specified path will be prepended to the filename in "new".   Provided
     the make_dir flag is enabled, any necessary directories to accomodate
     the new  filename  will be created. When a prefix is specified, a '^'
     in the received filename is not tolerated.

IMPORT:	old - Pointer to old filename.
EXPORT: new - Pointer to new filename.
		TRUE - Filename is valid.
		FALSE - Filename is not valid.
----------------------------------------------------------------------------*/
make_file_name( char *new, char *old ) {
	register char	*p;
	register int len = strlen( old );

	strcpy( new, prefix );					/* Set prefix into new name		*/
	if ( force ) return( TRUE );			/* If force flag, then all done	*/
	if ( !len ) return( FALSE );			/* Must be a file name			*/
	p = old;								/* Start with old name			*/
	if ( prefix[0] ) {		/* No directory climbing if a prefix was set	*/
		if ( strchr( old, '^' ) ) return( FALSE );	/* QNX 2.15 style		*/		p = old;
		while ( *p ) {
			if ( *p == '.'  &&  *(p+1) == '.' ) return( FALSE );
			++p;
			}
		p = old;					/* Start with old name					*/
		if ( *p == '[' )			/* Skip over the QNX 2.15 node number	*/
			while( *p  &&  *p++ != ']' ) ;
		if ( *p == '/'  &&  *(p+1) == '/' ) {
			p += 2;					/* Skip over the QNX4.0 node number		*/
			while( *p  &&  *p++ != '/' ) ;
			}
		if ( *p == '$' ) return( FALSE );	/* Can't prepend a device name	*/
		if ( *p && *(p+1) == ':') p += 2;	/* Skip over the drive number	*/
		if ( *p == '/' ) ++p;				/* Skip any path info			*/
		if ( *p == 0) return( FALSE );		/* Nothing left...				*/
		}
	strcat( new, p );	/* Place what's left of old name into the new name	*/
	if ( qnx_create_path_to( new ) == -1 ) return( FALSE );
	return( TRUE );
	}

/*	A useful debugging routine	*/

static char *stbl[33] = {
	"nul"  /* 0x00 */,	"soh"  /* 0x01 */,	"stx"  /* 0x02 */,
	"etx"  /* 0x03 */,	"eot"  /* 0x04 */,	"enq"  /* 0x05 */,
	"ack"  /* 0x06 */,	"bel"  /* 0x07 */,	"\\b " /* 0x08 */,
	"\\t " /* 0x09 */,	"\\n " /* 0x0a */,	"vt "  /* 0x0b */,
	"\\f " /* 0x0c */,	"\\r " /* 0x0d */,	"so "  /* 0x0e */,
	"si "  /* 0x0f */,	"dle"  /* 0x10 */,	"dc1"  /* 0x11 */,
	"dc2"  /* 0x12 */,	"dc3"  /* 0x13 */,	"dc4"  /* 0x14 */,
	"nak"  /* 0x15 */,	"syn"  /* 0x16 */,	"etb"  /* 0x17 */,
	"can"  /* 0x18 */,	"em "  /* 0x19 */,	"sub"  /* 0x1a */,
	"\\e " /* 0x1b */,	"fs "  /* 0x1c */,	"gs "  /* 0x1d */,
	"rs "  /* 0x1e */,	"us "  /* 0x1f */,	"sp "  /* 0x20 */
};

void
dump( char *msg, int n, uchar *buf ) {
	register int i;

	if (debug_flag) {
	    char line[80], *lp;
	    register uint_t field_width = 16;
	    register uchar_t *bptr = buf;
	    register uchar_t c;
	    fprintf( stderr, "%s, count = %d (%#x):\n", msg, n, n );
	    /*
	     * display two lines: one hex & one ascii
	     */
	    for (i = 0; i < n; i++) {
		if ((i % field_width) == 0) {
		    int cnt;
		    lp = line;
		    if (i) {
			fputs ("\n", stderr);
			fprintf (stderr, "%s\n", line);
		    }
#if defined(_QNX_SOURCE)
		    sprintf (lp, "%#x ", bptr);
#else /* !defined(_QNX_SOURCE) */
		    sprintf (lp, "%#lx ", bptr);
#endif /* defined(_QNX_SOURCE) */
		    cnt = strlen(lp);
		    fputs (lp, stderr);
		    sprintf(lp, "%*s", cnt, " ");
		    lp += cnt;
		}
		fprintf (stderr, " %02x ", *bptr);
		if ((c = *bptr++) <= SP) {
		    sprintf (lp, " %s", stbl[c]);
		} else if (c == DEL) {
		    sprintf (" %s", "del");
		} else if (c < DEL) {
		    sprintf (lp, "  %c ", c);
		} else {
		    sprintf (lp, "    ");
		}
		lp += strlen(lp);
	    }
	    if (i) {
		fputs ("\n", stderr);
		fprintf (stderr, "%s\n", line);
	    }
	    fflush( stderr );
	} else {
	    putmsg( "\r\012%s\ncount = %d\r\012 ", msg, n );
	    for ( i = 0; i < n; ++i ) putmsg( " %02x", buf[i] );
	    printf( "\r\012" );
	    fflush( stdout );
	}
}

void
flush_in() {
	tcflush( mdm, TCIFLUSH );
	}

void
proto_ungetc( uchar ch ) {
	ungot_char = (unsigned) ch;
	}

char *
fgetln( char *buf, int length, FILE *fp ) {
	register int len;

	if ( fgets( buf, length, fp ) != buf ) return( NULL );
	if ( (len = strlen( buf )) > 0 ) {
		if ( buf[ len-1 ] == '\n' ) buf[ len-1 ] = '\0';
		}
	return( buf );
	}

void
raw( int fd ) {
#if defined(_QNX_SOURCE)
	old_mode = dev_mode( fd, 0, _DEV_MODES );
#else /* !defined(_QNX_SOURCE) */
	(void) setup_tty( fd, (fd != mdm) ? &local_tty : &modem_tty );
#endif /* defined(_QNX_SOURCE) */
	}

void
restore_opt( int fd ) {
#if defined(_QNX_SOURCE)
	dev_mode( fd, old_mode, _DEV_MODES );
#else /* !defined(_QNX_SOURCE) */
	(void) restore_tty( fd, (fd != mdm) ? &local_tty : &modem_tty );
#endif /* defined(_QNX_SOURCE) */
	}

#if !defined(_QNX_SOURCE)

/************************************************************************
 *									*
 *	Functions added to support non-QNX Operating Systems.		*
 *									*
 ************************************************************************/

void
a_abrt(int err, char *fn, int fd)
{
	fprintf(stderr, "\r\nqcp: (%s) %s\r\n", fn, strerror(err));

	/*
	 * Restore the previous terminal characteristics.
	 */
	if (fd == mdm) (void) restore_tty (fd, &modem_tty);

	exit (EXIT_FAILURE);
}

ssize_t
a_read(int fildes, void *buf, size_t nbyte)
{
	ssize_t rc;	

	if ((rc = read(fildes,buf,nbyte))==-1) {
		switch (errno) {
			case EINTR:
			case EIO:		/* Old MacDonald had a farm, EI EIO... */
				break;
			default:		/* Old MacDonald bought the farm... */
				a_abrt(errno,"read",fildes);
		}
	}	

	return rc;
}


ssize_t
dev_read (int fd, void *buf, ssize_t n, unsigned minc, unsigned time,
				unsigned timeout, pid_t proxy, int *armed)
{
	fd_set	readmask;
	int	nfds, nf;
	struct	timeval timer;
	ssize_t	count;

#ifdef notdef
	if (debug_flag) {
	    putdbg("dev_read( int fd = %d, void *buf = %#lx, ssize_t n = %ld, unsigned minc = %d, unsigned time = %d, unsigned timeout = %d )\n",
				fd, buf, n, minc, time, timeout);
	}
#endif
	nfds = (fd + 1);
	timer.tv_sec = (timeout / onesec);
	timer.tv_usec = 0;
	FD_ZERO(&readmask);
	FD_SET(fd,&readmask), FD_SET(fd,&readmask);
	nf = select(nfds, &readmask, NULL, NULL, &timer);
	if ( (nf == FAILURE) && (errno != EINTR) ) {
	    a_abrt(errno, "select", fd);
	}
	if ( FD_ISSET(fd, &readmask) ) {
#ifdef notdef
	    /*
	     * This operation is corrupting input characters.  Why?
	     */
	    int status;
	    struct tty_params *tp = &modem_tty;
	    struct termios *tm= &tp->tty_cmodes;
	    if (tp->tty_saved &&
		(tm->c_cc[VMIN] != minc) || (tm->c_cc[VTIME] != time) ) {
		tm->c_cc[VMIN] = (minc < 256) ? minc : 0; /* Only 8 bits. */
		tm->c_cc[VTIME] = time;
		if ((status = tcsetattr (fd, TCSADRAIN, tm)) < 0) {
		    a_abrt (errno, "tcsetattr()", fd);
		}
	    }
#endif
	    count = a_read(fd, buf, n);
#ifdef notdef
	    if (debug_flag) {
		dump("data read", count, buf);
	    }
#endif
	    return (count);
	} else {
	    return (0);
	}
}


void
timerALRM(int val)
{
	longjmp(alrmbuf, 1);
}

/* Delay for specified microseconds. */

int
delay(int usecdelay)
{
	struct itimerval val, oval;
	int secdelay = 0;

	if(setjmp(alrmbuf)) {
		signal(SIGALRM, SIG_DFL);
		return(1);
	}
	oval.it_interval.tv_sec = 0;
	oval.it_interval.tv_usec = 0;
	oval.it_value.tv_sec = 0;
	oval.it_value.tv_usec = 0;
	val.it_interval.tv_sec = 0;
	val.it_interval.tv_usec = 0;
	/* make sure microseconds is less than 1 second */
	if (usecdelay >= 1000000) {
	    secdelay += usecdelay / 1000000;
	    usecdelay %= 1000000;
	}
	val.it_value.tv_sec = secdelay;
	val.it_value.tv_usec = usecdelay;
	signal(SIGALRM, timerALRM);
	setitimer(ITIMER_REAL, &val, &oval);
	pause();
	return(0);
}


/* Return number of characters to be read. */

int
dev_ischars(int fd)
{
	int status;
	int incnt = 0;

	if ((status = ioctl(fd, FIONREAD, (caddr_t)&incnt)) == FAILURE) {
		perror ("FIONREAD");
		return (status);
	} else {
		return (incnt);
	}
}


/************************************************************************
 *									*
 * save_tty() - Save The Current Terminal Characteristics.		*
 *									*
 * Inputs:	fd = The terminal file descriptor.			*
 *		tp = The tty parameters structure.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE.				*
 *									*
 ************************************************************************/
int
save_tty(int fd, struct tty_params *tp)
{
	int status;

	if (tp->tty_saved == TRUE) return(SUCCESS);

#if !defined(_QNX_SOURCE)
	/*
	 * Save the line discipline.
	 */
	if ((status = ioctl (fd, TIOCGETD, &tp->tty_ldisc)) < 0) {
	    a_abrt (errno, "TIOCGETD", fd);
	}
#endif /* !defined(_QNX_SOURCE) */

	/*
	 * Save the terminal characteristics.
	 */
	if ((status = tcgetattr (fd, &tp->tty_smodes)) < 0) {
	    a_abrt (errno, "tcgetattr()", fd);
	}
	if (status == SUCCESS) tp->tty_saved = TRUE;

	return (status);
}

/************************************************************************
 *									*
 * restore_tty() - Restore Saved Terminal Characteristics.		*
 *									*
 * Inputs:	fd = The terminal file descriptor.			*
 *		tp = The tty parameters structure.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE / WARNING (never saved)	*
 *									*
 ************************************************************************/
int
restore_tty(int fd, struct tty_params *tp)
{
	int status;

	if (tp->tty_saved == FALSE) return (WARNING);

	/*
	 * Restore the saved terminal characteristics.
	 */
	if ((status = tcsetattr (fd, TCSANOW, &tp->tty_smodes)) < 0) {
	    a_abrt (errno, "tcsetattr()", fd);
	}

#if !defined(_QNX_SOURCE)
	/*
	 * Restore the saved line discipline.
	 */
	if (ioctl (fd, TIOCSETD, &tp->tty_ldisc) < 0) {
	    a_abrt (errno, "TIOCSETD", fd);
	}
#endif /* !defined(_QNX_SOURCE) */
	return (status);
}

/************************************************************************
 *									*
 * setup_tty() - Setup local/modem terminal characteristics.		*
 *									*
 * Inputs:	fd = The terminal file descriptor.			*
 *		tp = The tty parameters structure.			*
 *									*
 * Return Value:							*
 *		Returns SUCCESS / FAILURE.				*
 *									*
 *   --> Presumes proper line discipline has already been setup. <--	*
 *									*
 ************************************************************************/
int
setup_tty(int fd, struct tty_params *tp)
{
	int status = SUCCESS;
	struct termios *tm = &tp->tty_cmodes;

	/*
	 * Save characteristics & setup initial defaults.
	 */
	if (tp->tty_saved == FALSE) {
		if ((status = save_tty (fd, tp)) == FAILURE) {
			return (status);
		}
	}
	*tm = *&tp->tty_smodes;			/* Copy saved modes. */

	/*
	 * Best Guess: Equivalent of 'stty raw' mode.
	 */
	/* Hardware Control Flags */
	tm->c_cflag &= ~(CSIZE | PARENB);	/* Init bits, no parity */
	tm->c_cflag |= CS8;			/* Ensure 8-bit data. */

	/* Input Flags */
	tm->c_iflag = 0;			/* No input processing. */

	/* Output Flags */
	tm->c_oflag = 0;			/* No output processing. */

	/* Local Flags */
	tm->c_lflag = 0;			/* Disable echo, etc. */

	/*
	 * Ensure these are set to normal default values.
	 */
	tm->c_cc[VMIN] = 1;
	tm->c_cc[VTIME] = 0;

	/*
	 * Setup the terminal characteristics now.
	 */
	if ((status = tcsetattr (fd, TCSANOW, tm)) < 0) {
	    a_abrt (errno, "tcsetattr()", fd);
	}

	return (status);
}

#define D(string)		fprintf (stderr, string)
#define D1(string, arg)		fprintf (stderr, string, arg)

void
print_usage( char *argv[] )
{
    D("qcp: [device] SEnd [options]* src_file[,dst_file] ... [x=index_file] ..\n");
    D("qcp: [device] REceive [options]* [-f forced_filename | -p prefix]\n");
    D("Options:\n");
    D(" -f filename           Force the received file to have this filename\n");
    D(" -l logfile            Log file transmission / receptions to this logfile\n");
    D(" -make_dir             Supress making directories for received files\n");
    D(" -newest               Receive only files newer than existing files\n");
    D(" -p prefix             Place this file or path prefix on received filenames\n");
    D(" -quiet                Be quiet during file transfer\n");
    D(" -relaxed_timing       Relaxed timing, double timeouts and quadruple timeouts\n");
    D(" -s packet_size        Specify an alternate packet size ( default: 2048 bytes )\n");
    D(" -today's_date         Stamp today's date on received files\n");
    D(" -unlink_file          Unlink existing files that lack write permission\n");
    D(" -Verbose, -verbose    Display error status while transferring files\n");
    D(" -D dbglogfile         Enable debugging output to specified log file.\n");
    exit (EXIT_FAILURE);
}

#endif /* !defined(_QNX_SOURCE) */
